"""
vqe_solver.py

This module contains a function to solve a quantum operator using the
Variational Quantum Eigensolver (VQE). The VQE algorithm combines quantum and
classical optimization to find the minimum eigenvalue of a Hamiltonian.
"""

import numpy as np
from qiskit.circuit.library import EfficientSU2
from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
from qiskit_aer.backends.aer_simulator import AerBackend
from qiskit_ibm_runtime import EstimatorV2, Session
from scipy.optimize import minimize

from quantum_pipeline.configs.module.backend import BackendConfig
from quantum_pipeline.solvers.solver import Solver
from quantum_pipeline.solvers.optimizer_config import get_optimizer_configuration
from quantum_pipeline.structures.vqe_observation import (
    VQEInitialData,
    VQEProcess,
    VQEResult,
)
from quantum_pipeline.utils.timer import Timer


class VQESolver(Solver):
    def __init__(
        self,
        qubit_op,
        backend_config: BackendConfig,
        max_iterations=50,
        optimizer='COBYLA',
        ansatz_reps=3,
        default_shots=1024,
        convergence_threshold=None,
        optimization_level=3,
    ):
        super().__init__()
        self.qubit_op = qubit_op
        self.ansatz_reps = ansatz_reps
        self.optimizer = optimizer
        self.max_iterations = max_iterations
        self.digits_iter = len(str(max_iterations))
        self.default_shots = default_shots
        self.backend_config = backend_config
        self.vqe_process: list[VQEProcess] = []
        self.current_iter = 1
        self.convergence_threshold = convergence_threshold
        self.optimization_level = optimization_level

    def _optimize_circuits(self, ansatz, hamiltonian, backend):
        """Prepare ISA-compatible circuits and observables"""
        target = backend.target
        pm = generate_preset_pass_manager(
            target=target, optimization_level=self.optimization_level
        )
        ansatz_isa = pm.run(ansatz)

        hamiltonian_isa = hamiltonian.apply_layout(layout=ansatz_isa.layout)
        return ansatz_isa, hamiltonian_isa

    def computeEnergy(self, params, ansatz, hamiltonian, estimator):
        """Return estimate of energy from estimator"""
        pub = (ansatz, [hamiltonian], [params])
        result = estimator.run(pubs=[pub]).result()
        energy, std = result[0].data.evs[0], result[0].data.stds[0]

        iter = VQEProcess(
            iteration=self.current_iter,
            parameters=params,
            result=energy,
            std=std,
        )

        self.vqe_process.append(iter)
        self.logger.debug(
            f'Iters. done: {self.current_iter:0{self.digits_iter}d} [Current cost: {energy}]'
        )
        self.current_iter += 1
        return energy

    def viaIBMQ(self, backend):
        """Run the VQE simulation on IBM Quantum backend."""
        hamiltonian = self.qubit_op

        self.logger.info(f'Initializing the ansatz with {self.ansatz_reps} reps...')
        ansatz = EfficientSU2(hamiltonian.num_qubits, reps=self.ansatz_reps)
        self.logger.info('Ansatz initialized.')

        param_num = ansatz.num_parameters
        x0 = 2 * np.pi * np.random.random(param_num)
        self.logger.debug(f'Initial ansatz parameters:\n\n{x0}\n')

        self.logger.info('Optimizing ansatz and hamiltonian...')
        ansatz_isa, hamiltonian_isa = self._optimize_circuits(ansatz, hamiltonian, backend)
        self.logger.info('Ansatz and hamiltonian optimized.')

        self.init_data = VQEInitialData(
            backend=backend.name,
            num_qubits=hamiltonian_isa.num_qubits,
            hamiltonian=hamiltonian_isa.to_list(),
            num_parameters=ansatz_isa.num_parameters,
            initial_parameters=x0,
            optimizer=self.optimizer,
            ansatz=ansatz_isa,
            ansatz_reps=self.ansatz_reps,
            noise_backend=self.backend_config.noise if self.backend_config.noise else 'undef',
            default_shots=self.default_shots,
        )
        self.logger.info('Opening a session...')

        with Session(backend=backend) as session:
            estimator = EstimatorV2(mode=session)
            estimator.options.default_shots = self.default_shots

            optimization_params, minimize_tol = get_optimizer_configuration(
                optimizer=self.optimizer,
                max_iterations=self.max_iterations,
                convergence_threshold=self.convergence_threshold,
                num_parameters=len(x0)
            )

            self.logger.debug(f'Optimization params: {optimization_params}')
            if minimize_tol is not None:
                self.logger.debug(f'Minimize tolerance: {minimize_tol}')

            if self.convergence_threshold and self.max_iterations:
                self.logger.info(
                    f'Starting VQE optimization with max iterations {self.max_iterations} taking priority over '
                    f'convergence threshold {self.convergence_threshold}'
                )
            elif self.convergence_threshold:
                self.logger.info(
                    f'Starting VQE optimization with convergence threshold {self.convergence_threshold}'
                )
            elif self.max_iterations:
                self.logger.info(
                    f'Starting VQE optimization with max iterations {self.max_iterations}'
                )
            else:
                self.logger.info(
                    f'Starting VQE optimization with default settings'
                )

        with Timer() as t:
            res = minimize(
                self.computeEnergy,
                x0,
                args=(ansatz_isa, hamiltonian_isa, estimator),
                method=self.optimizer,
                options=optimization_params,
                tol=minimize_tol,
            )

        actual_iterations = len(self.vqe_process)
        if self.convergence_threshold:
            if res.success:
                self.logger.info(
                    f'VQE converged after {actual_iterations} iterations (threshold: {self.convergence_threshold})'
                )
            else:
                self.logger.info(
                    f'VQE stopped after {actual_iterations} iterations - convergence not achieved'
                )
        else:
            self.logger.info(f'VQE completed {actual_iterations} iterations')

        result = VQEResult(
            initial_data=self.init_data,
            iteration_list=self.vqe_process,
            minimum=res.fun,
            optimal_parameters=res.x,
            maxcv=getattr(res, 'maxcv', None),
            minimization_time=np.float64(t.elapsed),
        )

        self.logger.info('Session closed.')
        self.logger.info(
            f'Calculations on quantum hardware via IBMQ completed in {t.elapsed:}.6f seconds.'
        )
        return result

    def viaAer(self, backend):
        """Run the VQE simulation via Aer simulator."""
        hamiltonian = self.qubit_op

        self.logger.info('Initializing the ansatz...')
        ansatz = EfficientSU2(hamiltonian.num_qubits)
        self.logger.info('Ansatz initialized.')

        param_num = ansatz.num_parameters
        x0 = 2 * np.pi * np.random.random(param_num)
        self.logger.debug(f'Initial ansatz parameters:\n\n{x0}\n')

        self.logger.info('Optimizing ansatz and hamiltonian...')
        ansatz_isa, hamiltonian_isa = self._optimize_circuits(ansatz, hamiltonian, backend)
        self.logger.info('Ansatz and hamiltonian optimized.')

        self.init_data = VQEInitialData(
            backend=backend.name,
            num_qubits=hamiltonian_isa.num_qubits,
            hamiltonian=hamiltonian_isa.to_list(),
            num_parameters=ansatz_isa.num_parameters,
            initial_parameters=x0,
            optimizer=self.optimizer,
            ansatz=ansatz_isa,
            ansatz_reps=self.ansatz_reps,
            noise_backend=self.backend_config.noise if self.backend_config.noise else 'undef',
            default_shots=self.default_shots,
        )

        estimator = EstimatorV2(mode=backend)
        estimator.options.default_shots = self.default_shots

        optimization_params, minimize_tol = get_optimizer_configuration(
            optimizer=self.optimizer,
            max_iterations=self.max_iterations,
            convergence_threshold=self.convergence_threshold,
            num_parameters=len(x0)
        )

        self.logger.debug(f'Optimization params: {optimization_params}')
        if minimize_tol is not None:
            self.logger.debug(f'Minimize tolerance: {minimize_tol}')

        if self.convergence_threshold and self.max_iterations:
            self.logger.info(
                f'Starting VQE optimization with max iterations {self.max_iterations} taking priority over '
                f'convergence threshold {self.convergence_threshold}'
            )
        elif self.convergence_threshold:
            self.logger.info(
                f'Starting VQE optimization with convergence threshold {self.convergence_threshold}'
            )
        elif self.max_iterations:
            self.logger.info(
                f'Starting VQE optimization with max iterations {self.max_iterations}'
            )
        else:
            self.logger.info(
                f'Starting VQE optimization with default settings'
            )
        with Timer() as t:
            res = minimize(
                self.computeEnergy,
                x0,
                args=(ansatz_isa, hamiltonian_isa, estimator),
                method=self.optimizer,
                options=optimization_params,
                tol=minimize_tol,
            )

        actual_iterations = len(self.vqe_process)
        if self.convergence_threshold:
            if res.success:
                self.logger.info(
                    f'VQE converged after {actual_iterations} iterations (threshold: {self.convergence_threshold})'
                )
            else:
                self.logger.info(
                    f'VQE stopped after {actual_iterations} iterations - convergence not achieved'
                )
        else:
            self.logger.info(f'VQE completed {actual_iterations} iterations')

        result = VQEResult(
            initial_data=self.init_data,
            iteration_list=self.vqe_process,
            minimum=res.fun,
            optimal_parameters=res.x,
            maxcv=getattr(res, 'maxcv', None),
            minimization_time=np.float64(t.elapsed),
        )

        self.logger.info(
            f'Simulation via Aer completed in {t.elapsed:.6f} seconds and {len(result.iteration_list)} iterations.'
        )
        return result

    def solve(self):
        """Run the VQE simulation and return the result."""
        self.current_iter = 1
        self.vqe_process = []

        backend = self.get_backend()

        result = None
        if isinstance(backend, AerBackend):
            result = self.viaAer(backend)
        else:
            result = self.viaIBMQ(backend)

        return result



from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from qiskit.quantum_info import SparsePauliOp
from qiskit_aer.backends.aer_simulator import AerBackend

from quantum_pipeline.configs.module.backend import BackendConfig
from quantum_pipeline.solvers.vqe_solver import VQESolver
from quantum_pipeline.structures.vqe_observation import VQEResult


@pytest.fixture
def mock_backend_config():
    """Create a mock BackendConfig for testing."""
    return BackendConfig(
        local=True,
        gpu=False,
        optimization_level=2,
        min_num_qubits=4,
        filters=None,
        simulation_method='mock_backend',
        gpu_opts=None,
        noise=None,
    )


@pytest.fixture
def sample_hamiltonian():
    """Create a sample Hamiltonian for testing."""
    return SparsePauliOp.from_list([('IIII', 0.1), ('IIXI', 0.2), ('XIXI', 0.3)])


@pytest.fixture
def vqe_solver(mock_backend_config, sample_hamiltonian):
    """Create a VQESolver instance for testing."""
    return VQESolver(
        qubit_op=sample_hamiltonian,
        backend_config=mock_backend_config,
        max_iterations=10,
        optimizer='COBYLA',
        ansatz_reps=2,
    )


def test_vqe_solver_initialization(vqe_solver, sample_hamiltonian, mock_backend_config):
    """Test VQESolver initialization."""
    assert vqe_solver.qubit_op == sample_hamiltonian
    assert vqe_solver.backend_config == mock_backend_config
    assert vqe_solver.max_iterations == 10
    assert vqe_solver.optimizer == 'COBYLA'
    assert vqe_solver.ansatz_reps == 2
    assert vqe_solver.current_iter == 1
    assert len(vqe_solver.vqe_process) == 0


def test_compute_energy(vqe_solver):
    """Test computeEnergy method."""
    # mocking the estimator
    mock_estimator = MagicMock()
    mock_result = MagicMock()
    mock_result.data.evs = [1.5]
    mock_result.data.stds = [0.1]
    mock_estimator.run.return_value.result.return_value = [mock_result]

    # mock the ansatz and hamiltonian
    mock_ansatz = MagicMock()
    mock_hamiltonian = MagicMock()

    # test params
    params = np.array([0.1, 0.2, 0.3])

    # suppress logging output
    with patch.object(vqe_solver.logger, 'debug'):
        energy = vqe_solver.computeEnergy(params, mock_ansatz, mock_hamiltonian, mock_estimator)

    assert energy == 1.5
    assert len(vqe_solver.vqe_process) == 1
    assert vqe_solver.vqe_process[0].iteration == 1
    assert vqe_solver.vqe_process[0].result == 1.5
    assert vqe_solver.current_iter == 2


def test_optimize_circuits(vqe_solver):
    """Test _optimize_circuits method."""
    # mock backend and its target
    mock_backend = MagicMock()
    mock_backend.target = MagicMock()

    # mock ansatz and hamiltonian
    mock_ansatz = MagicMock()
    mock_hamiltonian = MagicMock()

    # mock circuit optimization via pass_manager
    with patch('quantum_pipeline.solvers.vqe_solver.generate_preset_pass_manager') as mock_pm:
        # mock the pm instance
        mock_pass_manager = MagicMock()
        mock_pass_manager.run.return_value = mock_ansatz
        mock_pm.return_value = mock_pass_manager

        # mock the apply layout method
        mock_hamiltonian.apply_layout.return_value = mock_hamiltonian

        # call the method
        optimized_ansatz, optimized_hamiltonian = vqe_solver._optimize_circuits(
            mock_ansatz, mock_hamiltonian, mock_backend
        )

        mock_pm.assert_called_once_with(
            target=mock_backend.target, optimization_level=vqe_solver.optimization_level
        )
        assert optimized_ansatz == mock_ansatz
        mock_hamiltonian.apply_layout.assert_called_once()


def test_solve_via_aer(vqe_solver):
    """Test solve method with Aer backend."""
    mock_backend = MagicMock(spec=AerBackend)

    # patch the backend
    with (
        patch.object(vqe_solver, 'get_backend', return_value=mock_backend),
        patch.object(vqe_solver, 'viaAer') as mock_via_aer,
    ):
        # assign mock result
        mock_result = MagicMock()
        mock_via_aer.return_value = mock_result

        # call the solver
        result = vqe_solver.solve()

        vqe_solver.get_backend.assert_called_once()
        mock_via_aer.assert_called_once_with(mock_backend)
        assert result == mock_result


def test_solve_via_ibmq(vqe_solver):
    """Test solve method with IBMQ backend."""
    mock_backend = MagicMock()

    # patch the get_backend to run via ibm quantum
    with (
        patch.object(vqe_solver, 'get_backend', return_value=mock_backend),
        patch.object(vqe_solver, 'viaIBMQ') as mock_via_ibmq,
    ):
        # mock result
        mock_result = MagicMock()
        mock_via_ibmq.return_value = mock_result

        # call the solver
        result = vqe_solver.solve()

        vqe_solver.get_backend.assert_called_once()
        mock_via_ibmq.assert_called_once_with(mock_backend)
        assert result == mock_result


def test_vqe_solver_custom_parameters(mock_backend_config, sample_hamiltonian):
    """Test VQESolver with custom parameters."""
    custom_solver = VQESolver(
        qubit_op=sample_hamiltonian,
        backend_config=mock_backend_config,
        max_iterations=100,
        optimizer='SPSA',
        ansatz_reps=5,
        default_shots=2048,
        convergence_threshold=1e-4,
        optimization_level=2,
    )

    assert custom_solver.max_iterations == 100
    assert custom_solver.optimizer == 'SPSA'
    assert custom_solver.ansatz_reps == 5
    assert custom_solver.default_shots == 2048
    assert custom_solver.convergence_threshold == 1e-4
    assert custom_solver.optimization_level == 2


@pytest.mark.parametrize('backend_type', [AerBackend, MagicMock])
def test_solve_method_result_type(vqe_solver, backend_type):
    """Test that solve method returns a VQEResult for different backend types."""
    # mock backend
    mock_backend = MagicMock(spec=backend_type)
    mock_backend.name = 'test_backend'

    # mock result of the solver
    mock_result = MagicMock(spec=VQEResult)

    # patch get_backend to return the mock backend
    with patch.object(vqe_solver, 'get_backend', return_value=mock_backend):
        # patch method based on backend type
        if backend_type == AerBackend:
            with patch.object(vqe_solver, 'viaAer', return_value=mock_result):
                result = vqe_solver.solve()
        else:
            with patch.object(vqe_solver, 'viaIBMQ', return_value=mock_result):
                result = vqe_solver.solve()

        # verify if correct result was returned
        assert isinstance(result, VQEResult | MagicMock)
        assert result._spec_class == VQEResult


class TestVQEConvergencePriority:
    """Test suite for VQE convergence and max_iterations priority logic."""

    def test_max_iterations_priority_over_convergence(self):
        """Test that max_iterations takes priority when both are specified."""
        # Test the logic directly without full VQE execution
        max_iterations = 5
        convergence_threshold = 1e-6

        # Logic from VQE solver: tol should be None when both are specified
        # tol = convergence_threshold if convergence_threshold and not max_iterations else None
        tol = convergence_threshold if convergence_threshold and not max_iterations else None

        assert tol is None, "tol should be None when max_iterations takes priority over convergence"

    def test_convergence_only_uses_tolerance(self):
        """Test that convergence threshold is used when only convergence is specified."""
        max_iterations = None
        convergence_threshold = 1e-6

        # Logic from VQE solver
        tol = convergence_threshold if convergence_threshold and not max_iterations else None

        assert tol == 1e-6, "tol should be set when only convergence is specified"

    def test_max_iterations_only_no_tolerance(self):
        """Test that no tolerance is used when only max_iterations is specified."""
        max_iterations = 10
        convergence_threshold = None

        # Logic from VQE solver
        tol = convergence_threshold if convergence_threshold and not max_iterations else None

        assert tol is None, "tol should be None when only max_iterations is specified"

    def test_neither_specified_no_tolerance(self):
        """Test that no tolerance is used when neither is specified."""
        max_iterations = None
        convergence_threshold = None

        # Logic from VQE solver
        tol = convergence_threshold if convergence_threshold and not max_iterations else None

        assert tol is None, "tol should be None when neither is specified"

    def test_vqe_solver_initialization_with_both(self, sample_hamiltonian, mock_backend_config):
        """Test that VQESolver can be initialized with both parameters."""
        solver = VQESolver(
            qubit_op=sample_hamiltonian,
            backend_config=mock_backend_config,
            max_iterations=5,
            convergence_threshold=1e-6,
            optimizer='COBYLA',
            ansatz_reps=2,
        )

        assert solver.max_iterations == 5
        assert solver.convergence_threshold == 1e-6
        assert solver.optimizer == 'COBYLA'

    def test_vqe_solver_initialization_convergence_only(self, sample_hamiltonian, mock_backend_config):
        """Test that VQESolver can be initialized with only convergence threshold."""
        solver = VQESolver(
            qubit_op=sample_hamiltonian,
            backend_config=mock_backend_config,
            max_iterations=None,
            convergence_threshold=1e-6,
            optimizer='COBYLA',
            ansatz_reps=2,
        )

        assert solver.max_iterations is None
        assert solver.convergence_threshold == 1e-6

    def test_vqe_solver_initialization_max_iterations_only(self, sample_hamiltonian, mock_backend_config):
        """Test that VQESolver can be initialized with only max_iterations."""
        solver = VQESolver(
            qubit_op=sample_hamiltonian,
            backend_config=mock_backend_config,
            max_iterations=10,
            convergence_threshold=None,
            optimizer='COBYLA',
            ansatz_reps=2,
        )

        assert solver.max_iterations == 10
        assert solver.convergence_threshold is None

    def test_lbfgs_b_priority_logic(self):
        """Test L-BFGS-B specific logic when max_iterations is specified."""
        optimizer = 'L-BFGS-B'
        max_iterations = 5
        convergence_threshold = 1e-6

        # Test the specific L-BFGS-B logic - should disable early convergence when max_iterations is set
        should_disable_early_convergence = (
            optimizer == 'L-BFGS-B' and
            max_iterations
        )

        assert should_disable_early_convergence, "L-BFGS-B should disable default convergence when max_iterations is set"

        # Verify the ftol and gtol values that would be set
        expected_ftol = 1e-15
        expected_gtol = 1e-15

        assert expected_ftol < 1e-10, "ftol should be very small to prevent early convergence"
        assert expected_gtol < 1e-10, "gtol should be very small to prevent early convergence"

    def test_lbfgs_b_without_max_iterations(self):
        """Test L-BFGS-B behavior when max_iterations is not specified."""
        optimizer = 'L-BFGS-B'
        max_iterations = None

        # Should NOT disable early convergence when max_iterations is not set
        should_disable_early_convergence = (
            optimizer == 'L-BFGS-B' and
            max_iterations
        )

        assert not should_disable_early_convergence, "L-BFGS-B should use default convergence when max_iterations is not set"


class TestVQEEdgeCases:
    """Test edge cases and boundary conditions for VQE solver."""

    def test_max_iterations_zero(self, sample_hamiltonian, mock_backend_config):
        """Test VQESolver behavior with max_iterations=0."""
        solver = VQESolver(
            qubit_op=sample_hamiltonian,
            backend_config=mock_backend_config,
            max_iterations=0,
            optimizer='COBYLA',
            ansatz_reps=2,
        )
        assert solver.max_iterations == 0

    def test_max_iterations_negative(self, sample_hamiltonian, mock_backend_config):
        """Test VQESolver behavior with negative max_iterations."""
        solver = VQESolver(
            qubit_op=sample_hamiltonian,
            backend_config=mock_backend_config,
            max_iterations=-5,
            optimizer='COBYLA',
            ansatz_reps=2,
        )
        assert solver.max_iterations == -5

    def test_max_iterations_very_large(self, sample_hamiltonian, mock_backend_config):
        """Test VQESolver behavior with very large max_iterations."""
        large_value = 999999
        solver = VQESolver(
            qubit_op=sample_hamiltonian,
            backend_config=mock_backend_config,
            max_iterations=large_value,
            optimizer='L-BFGS-B',
            ansatz_reps=2,
        )
        assert solver.max_iterations == large_value

    def test_convergence_threshold_zero(self, sample_hamiltonian, mock_backend_config):
        """Test VQESolver behavior with convergence_threshold=0."""
        solver = VQESolver(
            qubit_op=sample_hamiltonian,
            backend_config=mock_backend_config,
            max_iterations=10,
            convergence_threshold=0.0,
            optimizer='L-BFGS-B',
            ansatz_reps=2,
        )
        assert solver.convergence_threshold == 0.0

    def test_convergence_threshold_negative(self, sample_hamiltonian, mock_backend_config):
        """Test VQESolver behavior with negative convergence_threshold."""
        solver = VQESolver(
            qubit_op=sample_hamiltonian,
            backend_config=mock_backend_config,
            max_iterations=10,
            convergence_threshold=-1e-6,
            optimizer='L-BFGS-B',
            ansatz_reps=2,
        )
        assert solver.convergence_threshold == -1e-6

    def test_convergence_threshold_very_small(self, sample_hamiltonian, mock_backend_config):
        """Test VQESolver behavior with very small convergence_threshold."""
        tiny_threshold = 1e-20
        solver = VQESolver(
            qubit_op=sample_hamiltonian,
            backend_config=mock_backend_config,
            max_iterations=10,
            convergence_threshold=tiny_threshold,
            optimizer='L-BFGS-B',
            ansatz_reps=2,
        )
        assert solver.convergence_threshold == tiny_threshold

    def test_lbfgs_b_with_max_iterations_one(self, sample_hamiltonian, mock_backend_config):
        """Test L-BFGS-B behavior with max_iterations=1."""
        solver = VQESolver(
            qubit_op=sample_hamiltonian,
            backend_config=mock_backend_config,
            max_iterations=1,
            optimizer='L-BFGS-B',
            ansatz_reps=2,
        )

        # Should still disable default convergence criteria
        should_disable = solver.optimizer == 'L-BFGS-B' and solver.max_iterations
        assert should_disable

    def test_cobyla_with_max_iterations_one(self, sample_hamiltonian, mock_backend_config):
        """Test COBYLA behavior with max_iterations=1."""
        solver = VQESolver(
            qubit_op=sample_hamiltonian,
            backend_config=mock_backend_config,
            max_iterations=1,
            optimizer='COBYLA',
            ansatz_reps=2,
        )

        # COBYLA should not have special convergence handling
        should_disable = solver.optimizer == 'L-BFGS-B' and solver.max_iterations
        assert not should_disable

    def test_different_optimizers_with_max_iterations(self, sample_hamiltonian, mock_backend_config):
        """Test that only L-BFGS-B gets special convergence handling."""
        optimizers = ['COBYLA', 'L-BFGS-B', 'COBYQA']

        for optimizer in optimizers:
            solver = VQESolver(
                qubit_op=sample_hamiltonian,
                backend_config=mock_backend_config,
                max_iterations=5,
                optimizer=optimizer,
                ansatz_reps=2,
            )

            should_disable = solver.optimizer == 'L-BFGS-B' and solver.max_iterations

            if optimizer == 'L-BFGS-B':
                assert should_disable, f"L-BFGS-B should disable convergence criteria"
            else:
                assert not should_disable, f"{optimizer} should not disable convergence criteria"

    def test_string_to_bool_edge_cases(self):
        """Test edge cases for convergence threshold truthiness."""
        # Test various falsy values for convergence_threshold
        falsy_values = [None, 0, 0.0, False]
        truthy_values = [1e-6, 1e-20, -1e-6, 1, True]

        for value in falsy_values:
            assert not bool(value), f"Value {value} should be falsy"

        for value in truthy_values:
            assert bool(value), f"Value {value} should be truthy"

    def test_optimization_params_structure(self):
        """Test that optimization parameters have correct structure."""
        max_iterations = 42

        # Base params that should always be present
        optimization_params = {
            'maxiter': max_iterations,
            'disp': False,
        }

        # L-BFGS-B specific additions
        if True:  # Simulating L-BFGS-B condition
            optimization_params.update({
                'ftol': 1e-15,
                'gtol': 1e-15,
            })

        assert 'maxiter' in optimization_params
        assert 'disp' in optimization_params
        assert optimization_params['maxiter'] == max_iterations
        assert optimization_params['disp'] is False
        assert optimization_params['ftol'] == 1e-15
        assert optimization_params['gtol'] == 1e-15



from unittest.mock import Mock, patch

from kafka.errors import KafkaError, NoBrokersAvailable
import pytest

from quantum_pipeline.configs.module.producer import ProducerConfig
from quantum_pipeline.configs.module.security import (
    CertConfig,
    SaslSslOpts,
    SecurityConfig,
)
from quantum_pipeline.stream.kafka_interface import KafkaProducerError, VQEKafkaProducer
from quantum_pipeline.structures.vqe_observation import VQEDecoratedResult


@pytest.fixture
def mock_config() -> ProducerConfig:
    """Create a mock ProducerConfig for testing."""
    mock_config = Mock(spec=ProducerConfig)
    mock_config.servers = ['localhost:9092']
    mock_config.topic = 'test_topic'
    mock_config.retries = 3
    mock_config.retry_delay = 1
    mock_config.kafka_retries = 3
    mock_config.acks = 'all'
    mock_config.timeout = 10

    mock_config.security = Mock(spec=SecurityConfig)
    mock_config.security.ssl = False
    mock_config.security.sasl_ssl = False
    mock_config.security.ssl_check_hostname = True

    mock_config.security.cert_config = Mock(CertConfig)
    mock_config.security.cert_config.ssl_dir = '/path/to/certs'
    mock_config.security.cert_config.ssl_cafile = 'ca.crt'
    mock_config.security.cert_config.ssl_certfile = 'client.crt'
    mock_config.security.cert_config.ssl_keyfile = 'client'
    mock_config.security.cert_config.ssl_password = 'testpass'
    mock_config.security.cert_config.ssl_crlfile = None
    mock_config.security.cert_config.ssl_ciphers = None

    mock_config.security.sasl_opts = Mock(SaslSslOpts)
    mock_config.security.sasl_opts.sasl_mechanism = 'PLAIN'
    mock_config.security.sasl_opts.sasl_plain_username = 'testuser'
    mock_config.security.sasl_opts.sasl_plain_password = 'testpass'
    mock_config.security.sasl_opts.sasl_kerberos_service_name = 'kerberos_name'
    mock_config.security.sasl_opts.sasl_kerberos_domain_name = 'kerberos_name'

    return mock_config


@pytest.fixture
def mock_result() -> VQEDecoratedResult:
    """Create a mock VQEDecoratedResult for testing."""
    mock_result = Mock(spec=VQEDecoratedResult)
    mock_result.get_schema_suffix.return_value = '_mol_test'
    return mock_result


class TestVQEKafkaProducer:
    def test_init_success(self, mock_config):
        """Test successful initialization of VQEKafkaProducer."""
        with patch('quantum_pipeline.stream.kafka_interface.VQEKafkaProducer') as mock_producer:
            mock_producer(mock_config)
            mock_producer.assert_called_once_with(mock_config)

    def test_init_no_brokers(self, mock_config):
        """Test handling of NoBrokersAvailable exception."""
        with patch(
            'quantum_pipeline.stream.kafka_interface.VQEKafkaProducer',
            side_effect=KafkaProducerError('No brokers available'),
        ) as mock_producer:
            with pytest.raises(KafkaProducerError, match='No brokers available'):
                mock_producer(mock_config)

    def test_security_config_ssl(self, mock_config):
        """Test SSL security configuration."""
        mock_config.security.ssl = True
        mock_config.security.cert_config.ssl_cafile = 'ca.crt'
        mock_config.security.cert_config.ssl_certfile = 'client.crt'
        mock_config.security.cert_config.ssl_keyfile = 'client.key'

        with patch('quantum_pipeline.stream.kafka_interface.VQEKafkaProducer') as mock_producer:
            mock_producer(mock_config)
            mock_producer.assert_called_once_with(mock_config)

            producer_config = mock_producer.call_args[0][0]

            assert producer_config.security.ssl
            assert producer_config.security.cert_config.ssl_cafile == 'ca.crt'
            assert producer_config.security.cert_config.ssl_certfile == 'client.crt'
            assert producer_config.security.cert_config.ssl_keyfile == 'client.key'

    def test_security_config_sasl_plain(self, mock_config):
        """Test SASL PLAIN authentication configuration."""
        mock_config.security.sasl_ssl = True
        mock_config.security.sasl_opts.sasl_mechanism = 'PLAIN'
        mock_config.security.sasl_opts.sasl_plain_username = 'testuser'
        mock_config.security.sasl_opts.sasl_plain_password = 'testpass'

        with patch('quantum_pipeline.stream.kafka_interface.VQEKafkaProducer') as mock_producer:
            mock_producer(mock_config)
            mock_producer.assert_called_once_with(mock_config)

            producer_config = mock_producer.call_args[0][0]

            assert producer_config.security.sasl_ssl
            assert producer_config.security.sasl_opts.sasl_mechanism == 'PLAIN'
            assert producer_config.security.sasl_opts.sasl_plain_username == 'testuser'
            assert producer_config.security.sasl_opts.sasl_plain_password == 'testpass'

    def test_invalid_sasl_config(self, mock_config):
        """Test handling of invalid SASL configuration."""
        mock_config.security.sasl_ssl = True
        mock_config.security.sasl_opts.sasl_mechanism = 'PLAIN'
        mock_config.security.sasl_opts.sasl_plain_username = None
        mock_config.security.sasl_opts.sasl_plain_password = None

        with pytest.raises(KafkaProducerError):
            VQEKafkaProducer(mock_config)

    def test_send_result_success(self, mock_config, mock_result):
        """Test successful sending of a result.

        This test verifies that:
            1. The producer is initialized correctly
            2. Serialization method is called
            3. send() and flush() methods are called on the Kafka producer
            4. Topic is updated before sending
        """

        try:
            # attempt to connect to Kafka, if unable - skip the test
            test_producer = VQEKafkaProducer(mock_config)
            test_producer.close()
        except NoBrokersAvailable:
            pytest.skip('Skipping test: No Kafka brokers available')
        except Exception as e:
            pytest.skip(f'Skipping test due to Kafka connection issue: {str(e)}')

        with patch(
            'quantum_pipeline.stream.kafka_security.KafkaSecurity',
            autospec=True,
        ) as mock_security_class:
            # mock security configuration
            mock_security = mock_security_class.return_value
            mock_security.build_security_config.return_value = {}

            with patch(
                'quantum_pipeline.utils.schema_registry.SchemaRegistry',
                autospec=True,
            ) as mock_registry_class:
                # create a mock schema registry class
                mock_registry_class.return_value

                # mock the producer class (along with internal kafka producer)
                producer = VQEKafkaProducer(mock_config)
                producer.producer = Mock()

                # mock metadata, returned by the producer after the send()
                mock_future = Mock()
                mock_metadata = Mock(topic='test_topic', partition=0, offset=1)
                mock_future.get.return_value = mock_metadata
                producer.producer.send.return_value = mock_future

                # mock serialization
                with patch.object(
                    producer, '_serialize_result', return_value=b'serialized_data'
                ) as mock_serialize:
                    # send result
                    producer.send_result(mock_result)

                    # check if the topic was updated
                    mock_result.get_schema_suffix.assert_called_once()

                    # check if serialization was triggered
                    mock_serialize.assert_called_once_with(mock_result)

                    # check if send was called with serialized data
                    producer.producer.send.assert_called_once_with(
                        mock_config.topic, b'serialized_data'
                    )

                    # check if flush was called
                    producer.producer.flush.assert_called_once()

    def test_send_result_failure(self, mock_config, mock_result):
        """Test handling of send failures.

        This test verifies that:
            1. An exception during sending raises a KafkaProducerError
            2. The topic is still updated before attempting to send
            3. The error contains appropriate context
        """
        try:
            # attempt to connect to Kafka, if unable - skip the test
            test_producer = VQEKafkaProducer(mock_config)
            test_producer.close()
        except NoBrokersAvailable:
            pytest.skip('Skipping test: No Kafka brokers available')
        except Exception as e:
            pytest.skip(f'Skipping test due to Kafka connection issue: {str(e)}')

        with patch(
            'quantum_pipeline.stream.kafka_security.KafkaSecurity',
            autospec=True,
        ) as mock_security_class:
            # mock security config
            mock_security = mock_security_class.return_value
            mock_security.build_security_config.return_value = {}

            # mock schema registry
            with patch(
                'quantum_pipeline.utils.schema_registry.SchemaRegistry',
                autospec=True,
            ):
                # create a real producer, with mock KafkaProducer
                producer = VQEKafkaProducer(mock_config)
                producer.producer = Mock()

                # mock serialization
                with patch.object(
                    producer, '_serialize_result', return_value=b'serialized_data'
                ) as mock_serialize:
                    # send sends an exception
                    producer.producer.send.side_effect = KafkaError('Send failed!')

                    # check if KafkaProducerError was raised
                    with pytest.raises(KafkaProducerError) as excinfo:
                        producer.send_result(mock_result)

                    # confirm that topic update was attempted
                    mock_result.get_schema_suffix.assert_called_once()

                    # confirm that serialization was attempted
                    mock_serialize.assert_called_once_with(mock_result)

                    # check that the error message contains the relevant context
                    assert 'Send failed!' in str(excinfo.value)

                    # ensure no flush - since send failed
                    producer.producer.flush.assert_not_called()

    def test_close(self, mock_config):
        """Test closing the producer.

        This test verifies that:
            1. The close method successfully calls the internal producer's close method
            2. Logging occurs during successful closure
            3. Proper error handling happens if close fails
        """
        try:
            # attempt to connect to Kafka, if unable - skip the test
            test_producer = VQEKafkaProducer(mock_config)
            test_producer.close()
        except NoBrokersAvailable:
            pytest.skip('Skipping test: No Kafka brokers available')
        except Exception as e:
            pytest.skip(f'Skipping test due to Kafka connection issue: {str(e)}')

        with patch(
            'quantum_pipeline.stream.kafka_security.KafkaSecurity',
            autospec=True,
        ) as mock_security_class:
            mock_security = mock_security_class.return_value
            mock_security.build_security_config.return_value = {}

            with patch(
                'quantum_pipeline.utils.schema_registry.SchemaRegistry',
                autospec=True,
            ):
                # create an actual wrapper object, but mock internal Producer
                producer = VQEKafkaProducer(mock_config)
                producer.producer = Mock()

                # close the producer
                producer.close()

                # check if the method really was called
                producer.producer.close.assert_called_once()

    def test_close_with_error(self, mock_config):
        """Test error handling during producer closure.

        This test verifies that:
            1. Errors during close are properly handled
            2. A KafkaProducerError is raised
            3. Error logging occurs
        """
        try:
            # attempt to connect to Kafka, if unable - skip the test
            test_producer = VQEKafkaProducer(mock_config)
            test_producer.close()
        except NoBrokersAvailable:
            pytest.skip('Skipping test: No Kafka brokers available')
        except Exception as e:
            pytest.skip(f'Skipping test due to Kafka connection issue: {str(e)}')

        with patch(
            'quantum_pipeline.stream.kafka_security.KafkaSecurity',
            autospec=True,
        ) as mock_security_class:
            mock_security = mock_security_class.return_value
            mock_security.build_security_config.return_value = {}

            with patch(
                'quantum_pipeline.utils.schema_registry.SchemaRegistry',
                autospec=True,
            ):
                # create the actual producer, with mock internal producer
                producer = VQEKafkaProducer(mock_config)
                producer.producer = Mock()

                # raise an exception when close is called
                producer.producer.close.side_effect = Exception('Closure failed')

                # check if the exception is properly handled
                with pytest.raises(KafkaProducerError, match='Failed to close producer'):
                    producer.close()
